/*
 *	Ask SCSI status uAskSCSIv
 *	written by GORRY.
 *	modifyed by Eriko.
 */


#include	<ctype.h>
#include	<stdio.h>
#include	<stdarg.h>
#include	<stdlib.h>
#include	<string.h>
#include	<sys/stat.h>
#include	<doslib.h>
#include	<iocslib.h>


/***************************************************************/


#define	MaxArgc			2	/* ID, LUN */

#define	REQUESTSENSESIZE	14
#define	MININQUIRYSIZE		36
#define	INQUIRYSIZE		256
#define	MODESENSESIZE		48

#define	SCSIRETRYMaxCount	2

#define	SCSI_REQCODE	2
#define	SCSI_REQCODE	2
#define	SCSI_BUSYCODE	8


int Verbose = 0;
int Debug = 0;
int MaxID = 7;
int MaxLUN = 0;

int Argc;
char* Argv[MaxArgc + 1];
char* ProcName;

int ReadyTWOSCSI;
int SCSIRomLevel;
int SCSIRETRYCount;
unsigned long SCSI_BLOCKSIZE;
unsigned char RequestSenseBuf[256];
struct READCAP ReadCap;


/***************************************************************/


static __inline int
TWOSCHK (void)
{
  register int reg_d0 __asm ("d0");
  register int reg_d1 __asm ("d1") = 0x1e;

  __asm __volatile ("moveq.l #__SCSIDRV,%0\n\t"
		    "trap #15"
			: "=d" (reg_d0)
			: "d" (reg_d1));

  return reg_d0;
}

static __inline int
MAXIDCHK (void)
{
  register int reg_d0 __asm ("d0");
  register int reg_d1 __asm ("d1") = 0x1d;

  __asm __volatile ("moveq.l #__SCSIDRV,%0\n\t"
		    "trap #15"
			: "=d" (reg_d0)
			: "d" (reg_d1));

  return reg_d0;
}

static __inline int
TW_CALL (int no, int id)
{
  register int reg_d0 __asm ("d0");
  register int reg_d1 __asm ("d1") = no;
  register int reg_d4 __asm ("d4") = id;

  __asm __volatile ("moveq.l #__SCSIDRV,%0\n\t"
		    "trap #15"
			: "=d" (reg_d0)
			: "d" (reg_d1), "d" (reg_d4)
			: "d1", "d2", "d3", "d4");

  return reg_d0;
}


/************************************************************************
 *									*
 *	wv\							*
 *									*
 ************************************************************************/

static volatile void
PrintHelp (void)
{
  printf (" SCSI ɐڑĂ@̏\܂B\n"
	  " Usage:	%s [Option] [ID [LUN]]\n"
	  " ID:	0`%dA邢͏ȗB\n"
	  " LUN:	0`%dA邢͏ȗB\n"
	  " Option:-Ds,c	ZN^ s  c ZN^̓e\܂B\n"
	  "	-RS	SCSI oXZbgsȂ܂B\n"
	  "	-V	o[{[Y[hB\n", Argv[0], MaxID, MaxID);

  exit (EXIT_FAILURE);
}




/***************************************************************/


int
printferr (const char* format, ...)
{
  va_list ap;
  int cnt;
  struct stat st;

  va_start (ap, format);
  cnt = vfprintf (stderr, format, ap);
  fstat (fileno (stdout), &st);
  if (S_IFREG == (st.st_mode & S_IFMT))
    vprintf (format, ap);
  va_end (ap);
  return cnt;
}


void
Err_Memory (void)
{
  fprintf (stderr, "܂B\n");
  exit (EXIT_FAILURE);
}

void
Err_Read (int id, int ret)
{
  printferr ("ID%d  READ 삵܂(%X)B\n", id & 0xffff, ret);
  exit (EXIT_FAILURE);
}

void
Err_Option (char* p)
{
  fprintf (stderr, "%s: %s: IvVُłB\n", ProcName, p);
  exit (EXIT_FAILURE);
}

void
Err_OverArg (char* p)
{
  fprintf (stderr, "%s: %s: ܂B\n", ProcName, p);
  exit (EXIT_FAILURE);
}


/***************************************************************/

static void
WaitSec (int sec)
{
  int t1;
  int t2;

  t1 = ONTIME ();
  do
    {
      t2 = ONTIME ();
      if (t1 > t2)
	t2 += 8640000;
    }
  while (t1 + sec * 10 > t2);
}

void
ResetBus (void)
{
  if (Debug || Verbose)
    printf ("Reset SCSI Bus:\n");

  S_RESET ();
}

int
RequestSense (int id, void* buf)
{
  int ret;
  int i;

  if (Debug || Verbose)
    printf ("Read RequestSense:");

  ret = S_REQUEST (REQUESTSENSESIZE, id, buf);
  if (Debug || Verbose)
    printf ("%08X\n", ret);

  if (Debug)
    {
      if (!ret)
	{
	  for (i = 0; i < REQUESTSENSESIZE; i++)
	    {
	      printf ("%02X,", *((unsigned char*) (buf) + i));
	      if ((i % 16) == 15)
		printf ("\n");
	    }
	  printf ("\n");
	}
      else
	printf ("Request Sense f[^݂͑܂(%d)B\n", ret);
    }

  return ret;
}

int
StartStop (int id, int md)
{
  int ret;

  SCSIRETRYCount = 0;
Retry:;
  if (Debug || Verbose)
    printf ("StartStop(md=%d):", md);

  ret = S_STARTSTOP (id, md);
  if (Debug || Verbose)
    printf ("%08X\n", ret);

  if (ret > 0)
    {
      SCSIRETRYCount++;
      if (SCSIRETRYCount < SCSIRETRYMaxCount)
	{
	  switch (ret)
	    {
	    case SCSI_REQCODE:
	      RequestSense (id, RequestSenseBuf);
	      switch (RequestSenseBuf[2] & 0x0f)
		{
		case 2:	/* Not Ready */
		case 3:	/* Medium Error */
		case 4:	/* Hardware Error */
		case 6:	/* Unit Attention */
		  WaitSec (1);
		  goto Retry;
		}
	      break;
	    case SCSI_BUSYCODE:
	      WaitSec (1);
	      goto Retry;
	    }
	}
    }

  return ret;
}

int
RezeroUnit (int id)
{
  int ret;

  SCSIRETRYCount = 0;
Retry:;
  if (Debug || Verbose)
    printf ("Rezero Unit:");

  ret = S_REZEROUNIT (id);
  if (Debug || Verbose)
    printf ("%08X\n", ret);

  if (ret > 0)
    {
      SCSIRETRYCount++;
      if (SCSIRETRYCount < SCSIRETRYMaxCount)
	{
	  switch (ret)
	    {
	    case SCSI_REQCODE:
	      RequestSense (id, RequestSenseBuf);
	      switch (RequestSenseBuf[2] & 0x0f)
		{
		case 2:	/* Not Ready */
		case 3:	/* Medium Error */
		case 4:	/* Hardware Error */
		case 6:	/* Unit Attention */
		  WaitSec (1);
		  goto Retry;
		}
	      break;
	    case SCSI_BUSYCODE:
	      WaitSec (1);
	      goto Retry;
	    }
	}
    }

  return ret;
}

int
ModeSense (int id, int page, void* buf)
{
  register int reg_d0 __asm ("d0");
  int ret;
  int i;

  SCSIRETRYCount = 0;
Retry:;
  if (Debug || Verbose)
    printf ("Read ModeSense(page=%d):", page);

  __asm __volatile ("	move.l	%1,a1			\n"
		    "	moveq.l	#$F5,d0			\n"
		    "	moveq.l	#$29,d1			\n"
		    "	move.l	%3,d2			\n"
		    "	move.l	%4,d3			\n"
		    "	move.l	%2,d4			\n"
		    "	move.l	(_SCSI_BLOCKSIZE),d5	\n"
		    "	trap	#15			\n"
		: "=d" (reg_d0)
		: "g" (buf), "ri" ((int) id), "ri" ((int) page), "ri" (MODESENSESIZE)
		: "d1", "d2", "d3", "d4", "d5", "a1");
  ret = reg_d0;
  if (Debug || Verbose)
    printf ("%08X\n", ret);

  if (ret > 0)
    {
      SCSIRETRYCount++;
      if (SCSIRETRYCount < SCSIRETRYMaxCount)
	{
	  switch (ret)
	    {
	    case SCSI_REQCODE:
	      RequestSense (id, RequestSenseBuf);
	      switch (RequestSenseBuf[2] & 0x0f)
		{
		case 2:	/* Not Ready */
		case 3:	/* Medium Error */
		case 4:	/* Hardware Error */
		case 6:	/* Unit Attention */
		  WaitSec (1);
		  goto Retry;
		}
	      break;
	    case SCSI_BUSYCODE:
	      WaitSec (1);
	      goto Retry;
	    }
	}
    }

  if (Debug)
    {
      if (!ret)
	{
	  for (i = 0; i < MODESENSESIZE; i++)
	    {
	      printf ("%02X,", *((unsigned char*) (buf) + i));
	      if ((i % 16) == 15)
		printf ("\n");
	    }
	  printf ("\n");
	}
      else
	printf ("Mode Sense f[^݂͑܂(%d)B\n", ret);
    }

  return ret;
}


int
Inquiry (int id, void* buf)
{
  int ret;
  int i;
  int len;
  unsigned char c;

  SCSIRETRYCount = 0;
Retry:;
  if (Debug || Verbose)
    printf ("Read Inquiry:");

  len = MININQUIRYSIZE;
  ret = S_INQUIRY (len, id, buf);
  if (!ret)
    {
      len = 5 + (((char*) buf)[4]);
      if (len > INQUIRYSIZE)
	len = INQUIRYSIZE;
      ret = S_INQUIRY (len, id, buf);
    }
  if (Debug || Verbose)
    printf ("%08X\n", ret);

  if (ret > 0)
    {
      SCSIRETRYCount++;
      if (SCSIRETRYCount < SCSIRETRYMaxCount)
	{
	  switch (ret)
	    {
	    case SCSI_REQCODE:
	      RequestSense (id, RequestSenseBuf);
	      switch (RequestSenseBuf[2] & 0x0f)
		{
		case 2:	/* Not Ready */
		case 3:	/* Medium Error */
		case 4:	/* Hardware Error */
		case 6:	/* Unit Attention */
		  WaitSec (1);
		  goto Retry;
		}
	      break;
	    case SCSI_BUSYCODE:
	      WaitSec (1);
	      goto Retry;
	    }
	}
    }

  if (Debug)
    {
      if (!ret)
	{
	  printf ("len=%X:\n", len);
	  for (i = 0; i < len; i++)
	    {
	      printf ("%02X,", *((unsigned char*) (buf) + i));
	      if ((i % 16) == 15)
		printf ("\n");
	    }
	  printf ("\n");
	}
    }
  for (i = 8; i < 8 + 42; i++)
    {
      c = *((unsigned char*) (buf) + i);
      if (c < 0x20)
	break;
      if (c > 0x7f)
	break;
    }
  *((unsigned char*) (buf) + i) = '\0';

  return ret;
}


int
ReadCapacity (int id, void* buf)
{
  int ret;
  int i;

  SCSIRETRYCount = 0;
Retry:;
  if (Debug || Verbose)
    printf ("Read Capacity:");

  ret = S_READCAP (id, buf);
  if (Debug || Verbose)
    printf ("%08X\n", ret);

  if (ret > 0)
    {
      SCSIRETRYCount++;
      if (SCSIRETRYCount < SCSIRETRYMaxCount)
	{
	  switch (ret)
	    {
	    case SCSI_REQCODE:
	      RequestSense (id, RequestSenseBuf);
	      switch (RequestSenseBuf[2] & 0x0f)
		{
		case 2:	/* Not Ready */
		case 3:	/* Medium Error */
		case 4:	/* Hardware Error */
		case 6:	/* Unit Attention */
		  WaitSec (1);
		  goto Retry;
		}
	      break;
	    case SCSI_BUSYCODE:
	      WaitSec (1);
	      goto Retry;
	    }
	}
    }

  if (Debug)
    {
      if (!ret)
	{
	  for (i = 0; i < 8; i++)
	    printf ("%02X,", *((unsigned char*) (buf) + i));
	  printf ("\n");
	}
    }

  return ret;
}


int
ReadSec (int id, int pos, int size, void* buf)
{
  int ret;
  int i;
  int j;

  SCSIRETRYCount = 0;
Retry:;
  if (Debug || Verbose)
    printf ("ReadSec(pos=$%08X,size=$%04X):", pos, size);
  if (ReadCap.block >= 0x200000)
    ret = S_READEXT (pos, size, id, SCSI_BLOCKSIZE, buf);
  else
    ret = S_READ (pos, size, id, SCSI_BLOCKSIZE, buf);
  if (Debug || Verbose)
    printf ("%08X\n", ret);

  if (ret > 0)
    {
      SCSIRETRYCount++;
      if (SCSIRETRYCount < SCSIRETRYMaxCount)
	{
	  switch (ret)
	    {
	    case SCSI_REQCODE:
	      RequestSense (id, RequestSenseBuf);
	      switch (RequestSenseBuf[2] & 0x0f)
		{
		case 2:	/* Not Ready */
		case 3:	/* Medium Error */
		case 4:	/* Hardware Error */
		case 6:	/* Unit Attention */
		  WaitSec (1);
		  goto Retry;
		}
	      break;
	    case SCSI_BUSYCODE:
	      WaitSec (1);
	      goto Retry;
	    }
	}
    }

  if (Debug)
    {
      if (!ret)
	{
	  for (j = 0; j < size; j++)
	    {
	      printf ("Sec: $%08X\n", pos + j);
	      for (i = 0; i < ReadCap.size; i++)
		{
		  printf ("%02X,", *((unsigned char*) (buf) + i));
		  if ((i % 16) == 15)
		    printf ("\n");
		}
	    }
	  printf ("\n");
	}
    }

  return ret;
}


int
TestUnit (int id)
{
  int ret;

  SCSIRETRYCount = 0;
Retry:;
  if (Debug || Verbose)
    printf ("Test Unit:");
  ret = S_TESTUNIT (id);
  if (Debug || Verbose)
    printf ("%08X\n", ret);

  if (ret > 0)
    {
      SCSIRETRYCount++;
      if (SCSIRETRYCount < SCSIRETRYMaxCount)
	{
	  switch (ret)
	    {
	    case SCSI_REQCODE:
	      RequestSense (id, RequestSenseBuf);
	      switch (RequestSenseBuf[2] & 0x0f)
		{
		case 2:	/* Not Ready */
		case 3:	/* Medium Error */
		case 4:	/* Hardware Error */
		case 6:	/* Unit Attention */
		  WaitSec (1);
		  goto Retry;
		}
	      break;
	    case SCSI_BUSYCODE:
	      WaitSec (1);
	      goto Retry;
	    }
	}
    }

  return ret;
}


/***************************************************************/


unsigned long Buffers;
unsigned char InquiryBuf[INQUIRYSIZE];
unsigned char ModeSense3Buf[MODESENSESIZE];
unsigned char ModeSense4Buf[MODESENSESIZE];
unsigned char* BootSecBuf;
unsigned char* PartitionSecBuf;
unsigned char* SecBuf;
unsigned char InitiatorID;
unsigned long PartitionCnt;
unsigned long IBMFDISKPartSec;

struct MEDIA
{
  unsigned long lblock;
  unsigned long lbsize;
  unsigned long cylsize;
  unsigned long track;
  unsigned long head;
  char mediatype;
  char okmodesense;
  char needSUSIE98COption;
} Media;

struct PARTITION
{
  char OK;
  char HumanOK;
  char BootFlag;
  char ActiveFlag;
  unsigned long Start;
  unsigned long End;
  unsigned long Size;
  char Name[128];
  unsigned long SecSize;
  unsigned long CluSec;
  unsigned long FATSec;
  unsigned long FATCnt;
  unsigned long RootDir;
  unsigned long MaxSec;
  unsigned long MediaID;
  unsigned long FATSiz;
  char OEMName[8 + 1];
  char FATName[8 + 1];
} Partition[32];

char* DeviceName[0x20] = {
  /* 00 */ "Direct Access Device",
  /* 01 */ "Sequencial Access Device",
  /* 02 */ "Printer Device",
  /* 03 */ "Processor Device",
  /* 04 */ "W-Once,R-Multiple Device",
  /* 05 */ "CD-ROM Device",
  /* 06 */ "Scanner Device",
  /* 07 */ "Optical Memory Device",
  /* 08 */ "Medium Changer Device",
  /* 09 */ "Communication Device",
  /* 0A */ "Defined by ASC IT8:0A",
  /* 0B */ "Defined by ASC IT8:0B",
  /* 0C */ "Array Controller Device",
  /* 0D */ "Enclosure Service Device",
  /* 0E */ "Simplified Direct Access",
  /* 0F */ "Optical Card Device",
  /* 10 */ "Bridge Expander Device",
  /* 11 */ "(unsupported $11)",
  /* 12 */ "(unsupported $12)",
  /* 13 */ "(unsupported $13)",
  /* 14 */ "(unsupported $14)",
  /* 15 */ "(unsupported $15)",
  /* 16 */ "(unsupported $16)",
  /* 17 */ "(unsupported $17)",
  /* 18 */ "(unsupported $18)",
  /* 19 */ "(unsupported $19)",
  /* 1A */ "(unsupported $1A)",
  /* 1B */ "(unsupported $1B)",
  /* 1C */ "(unsupported $1C)",
  /* 1D */ "(unsupported $1D)",
  /* 1E */ "(unsupported $1E)",
  /* 1F */ "Unknown or No Device Type"
};

char* ANSIVersion[0x08] = {
  /* 00 */ "(unsupported $00)",
  /* 01 */ "SCSI",
  /* 02 */ "SCSI-2",
  /* 03 */ "SCSI-3",
  /* 04 */ "SPC-2",
  /* 05 */ "(unsupported $05)",
  /* 06 */ "(unsupported $06)",
  /* 07 */ "(unsupported $07)"
};

char* ResType[0x10] = {
  /* 00 */ "SCSI",
  /* 01 */ "CCS",
  /* 02 */ "SCSI-2",
  /* 03 */ "SCSI-3",
  /* 04 */ "(unsupported $04)",
  /* 05 */ "(unsupported $05)",
  /* 06 */ "(unsupported $06)",
  /* 07 */ "(unsupported $07)",
  /* 08 */ "(unsupported $08)",
  /* 09 */ "(unsupported $09)",
  /* 0A */ "(unsupported $0A)",
  /* 0B */ "(unsupported $0B)",
  /* 0C */ "(unsupported $0C)",
  /* 0D */ "(unsupported $0D)",
  /* 0E */ "(unsupported $0E)",
  /* 0F */ "(unsupported $0F)"
};

int DumpData = 0;
unsigned long DumpStart;
unsigned long DumpCount;

int ReqResetBus = 0;

int CPUNo;
int VERNo;
int SCSIPort;


/***************************************************************/


void
SetResetBus (void)
{
  ResetBus ();
}


int
SetStartStop (int id)
{
  return StartStop (id, 1);
}


int
SetRezeroUnit (int id)
{
  return RezeroUnit (id);
}


int
GetTestUnit (int id)
{
  return TestUnit (id);
}


int
GetInquiry (int id)
{
  memset (InquiryBuf, 0, INQUIRYSIZE);
  return Inquiry (id, &InquiryBuf[0]);
}


int
GetReadCapacity (int id)
{
  ReadCap.size = 0;
  ReadCap.block = 0;
  return ReadCapacity (id, &ReadCap);
}


int
GetModeSense3 (int id)
{
  memset (ModeSense3Buf, 0, MODESENSESIZE);
  return ModeSense (id, 3, &ModeSense3Buf[0]);
}


int
GetModeSense4 (int id)
{
  memset (ModeSense4Buf, 0, MODESENSESIZE);
  return ModeSense (id, 4, &ModeSense4Buf[0]);
}


void
GetBootSec (int id)
{
  int ret;

  BootSecBuf = malloc (Media.lbsize);
  if (NULL == BootSecBuf)
    Err_Memory ();
  if (Debug || Verbose)
    printf ("[Read BootSec] ");
  ret = ReadSec (id, 0, 1, BootSecBuf);
  if (ret)
    Err_Read (id, ret);
}


void
GetBootSecCD (int id)
{
  int ret;

  BootSecBuf = malloc (Media.lbsize);
  if (NULL == BootSecBuf)
    Err_Memory ();
  BootSecBuf[0] = 0;
  if (Debug || Verbose)
    printf ("[Read BootSec] ");
  ret = ReadSec (id, 16, 1, BootSecBuf);
  if (ret)
    BootSecBuf[0] = 0;
}


/***************************************************************/


void
GetMediaStatus (int id)
{
  int ret;
  unsigned long lblock;

  Media.lblock = ReadCap.block + 1;
  Media.lbsize = ReadCap.size;
  switch (ReadCap.size)
    {
    case 0:
      break;
    case 0x100:
      SCSI_BLOCKSIZE = 0;
      break;
    case 0x200:
      SCSI_BLOCKSIZE = 1;
      break;
    case 0x400:
      SCSI_BLOCKSIZE = 2;
      break;
    case 0x800:
      SCSI_BLOCKSIZE = 3;
      break;
    default:
      printferr ("ReadCapacity ̋Alُł(%lX)B\n", ReadCap.size);
      exit (EXIT_FAILURE);
    }

  ret = GetModeSense3 (id);
  ret = GetModeSense4 (id);
  if (ret == 0)
    {
      lblock = ((((long) ModeSense4Buf[0x04]) << 24)
	      + (((long) ModeSense4Buf[0x05]) << 16)
	      + (((long) ModeSense4Buf[0x06]) << 8)
	      + (((long) ModeSense4Buf[0x07]) << 0));
      if (lblock == 0)
	lblock = Media.lblock;

      Media.okmodesense = !0;
      Media.cylsize = lblock /
	(  (((long) ModeSense4Buf[0x0e]) << 16)
	 + (((long) ModeSense4Buf[0x0f]) << 8)
	 + (((long) ModeSense4Buf[0x10]) << 0));
      Media.track = ((((long) ModeSense3Buf[0x16]) << 8)
		  +  (((long) ModeSense3Buf[0x17]) << 0));
      Media.head = (((long) ModeSense4Buf[0x11]) << 0);
    }
  else
    {
      Media.okmodesense = 0;
      Media.cylsize = 0x200;
      Media.track = 0x200;
      Media.head = 1;
    }
}


/***************************************************************/


unsigned long
SearchPartitionHeadSec_PC98 (int id)
{
  int ret;
  int i, j;
  unsigned long block;
  unsigned char* secbuf;

  Media.needSUSIE98COption = !0;

  if (Debug || Verbose)
    printf ("[Search PartitionHeadSec]\n");

  secbuf = malloc (Media.lbsize * 0x10);
  if (NULL == secbuf)
    Err_Memory ();

  for (i = 1; i < 0x100; i++)
    {
      block = i * 0x10;
      ret = ReadSec (id, block, 0x10, secbuf);
      if (ret)
	Err_Read (id, ret);
      for (j = 0; j < 0x10; j++)
	{
	  if (Debug || Verbose)
	    printf ("%d\r", i * 0x10 + j);
	  if (secbuf[Media.lbsize * j] == 0xeb)
	    goto Found;
	}
    }

/* NotFound:; */
  free (secbuf);
  return 0;

Found:;
  free (secbuf);
  return (i * 0x10 + j);
}

void
GetPartitionSec_PC98 (int id)
{
  int ret;

  PartitionSecBuf = malloc (Media.lbsize);
  if (NULL == PartitionSecBuf)
    Err_Memory ();

  if (Debug || Verbose)
    printf ("[Read PartitionSec] ");
  ret = ReadSec (id, 1, 1, PartitionSecBuf);
  if (ret)
    Err_Read (id, ret);
}


void
GetDiskStatus_PC98 (int id)
{
  int index;
  int i;
  int ret;
  unsigned long l;

  SecBuf = malloc (Media.lbsize);
  if (NULL == SecBuf)
    Err_Memory ();

  GetPartitionSec_PC98 (id);

  printf ("Format		: PC-9801\n");
  for (index = 0; index < Media.lbsize; index += 0x20)
    {
      struct PARTITION* p = &Partition[PartitionCnt++];
      int RetryCnt;

      if (!PartitionSecBuf[index + 0x10])
	continue;

      for (i = 0; i < 16; i++)
	p->Name[i] = PartitionSecBuf[index + 0x10 + i];
      p->Name[i] = '\0';

      p->BootFlag =
	(((PartitionSecBuf[index + 0x00] & 0xa0) == 0x20) ? 0 : !0);
      p->ActiveFlag = ((PartitionSecBuf[index + 0x01] & 0x80) ? !0 : 0);
      RetryCnt = 0;
    Retry_cylsize:;
      p->Start = ((((unsigned long) PartitionSecBuf[index + 0x0b]) << 8)
		+ (((unsigned long) PartitionSecBuf[index + 0x0a]) << 0))
							* Media.cylsize;
      p->End =
	((((unsigned long) PartitionSecBuf[index + 0x0f]) << 8) +
	 (((unsigned long) PartitionSecBuf[index + 0x0e]) << 0) + 1)
						* Media.cylsize - 1;
      p->Size = ((p->End) - (p->Start) + 1) / 1024 * Media.lbsize / 1024;

      p->OK = !0;
#if 0
      if (memcmp ("MS-DOS", &PartitionSecBuf[index + 0x10], 6))
	continue;
#endif
      if (!(PartitionSecBuf[index + 0x00] & 0x20))
	continue;

      if (Debug || Verbose)
	printf ("[Read PartitionHeadSec] ");

      ret = ReadSec (id, p->Start, 1, SecBuf);
      if (ret)
	Err_Read (id, ret);
      if (SecBuf[0] != 0xeb)
	{
	  if (index == 0)
	    {
	      switch (RetryCnt)
		{
		case 0:
		  RetryCnt++;
		  if (((Media.cylsize) != (Media.track * Media.head)) &&
		      (Media.track * Media.head != 0))
		    {
		      Media.cylsize = Media.track * Media.head;
		      goto Retry_cylsize;
		    }
		  /* fall down */
		case 1:
		  RetryCnt++;
		  Media.cylsize = SearchPartitionHeadSec_PC98 (id);
		  if (Media.cylsize != 0)
		    goto Retry_cylsize;
		  break;
		}
	    }
	  continue;
	}
      l = (((unsigned long) SecBuf[12]) << 8)
	+ (((unsigned long) SecBuf[11]) << 0);
      if (Buffers < l)
	Buffers = l;
      p->SecSize = l;
      l = (((unsigned long) SecBuf[13]) << 0);
      p->CluSec = l;
      l = (((unsigned long) SecBuf[15]) << 8)
	+ (((unsigned long) SecBuf[14]) << 0);
      p->FATSec = l;
      l = (((unsigned long) SecBuf[16]) << 0);
      p->FATCnt = l;
      l = (((unsigned long) SecBuf[18]) << 8)
	+ (((unsigned long) SecBuf[17]) << 0);
      p->RootDir = l;
      l = (((unsigned long) SecBuf[20]) << 8)
	+ (((unsigned long) SecBuf[19]) << 0);
      if (l == 0)
	{
	  l = (((unsigned long) SecBuf[35]) << 24)
	    + (((unsigned long) SecBuf[34]) << 16)
	    + (((unsigned long) SecBuf[33]) << 8)
	    + (((unsigned long) SecBuf[32]) << 0);
	}
      p->MaxSec = l;
      l = (((unsigned long) SecBuf[21]) << 0);
      p->MediaID = l;
      l = (((unsigned long) SecBuf[23]) << 8)
	+ (((unsigned long) SecBuf[22]) << 0);
      p->FATSiz = l;
      p->HumanOK = !0;
    }
}


/***************************************************************/


void
GetPartitionSec_X68K (int id)
{
  int ret;

  PartitionSecBuf = malloc (Media.lbsize);
  if (NULL == PartitionSecBuf)
    Err_Memory ();

  if (Debug || Verbose)
    printf ("[Read PartitionSec] ");
  ret = ReadSec (id, 2048 / Media.lbsize, 1, PartitionSecBuf);
  if (ret)
    Err_Read (id, ret);
}


void
GetDiskStatus_X68K (int id)
{
  int index;
  int i;
  int ret;
  unsigned long l;

  SecBuf = malloc (Media.lbsize);
  if (NULL == SecBuf)
    Err_Memory ();

  GetPartitionSec_X68K (id);

  if (memcmp ("X68K", &PartitionSecBuf[0], 4))
    goto Quit;

  printf ("Format		: X68000 (%s)\n", &BootSecBuf[0x10]);

  for (index = 0x10; index < 0x100; index += 0x10)
    {
      struct PARTITION* p = &Partition[PartitionCnt++];

      if (!PartitionSecBuf[index])
	continue;

      for (i = 0; i < 8; i++)
	p->Name[i] = PartitionSecBuf[index + i];
      p->Name[i] = '\0';

      p->BootFlag = ((PartitionSecBuf[index + 0x08] & 0x02) ? 0 : !0);
      p->ActiveFlag = ((PartitionSecBuf[index + 0x08] & 0x01) ? 0 : !0);
      p->Start = ((((unsigned long) PartitionSecBuf[index + 0x09]) << 16)
		+ (((unsigned long) PartitionSecBuf[index + 0x0a]) << 8)
		+ (((unsigned long) PartitionSecBuf[index + 0x0b]) << 0))
						* 1024 / Media.lbsize;
      p->End = ((((unsigned long) PartitionSecBuf[index + 0x0d]) << 16)
	      + (((unsigned long) PartitionSecBuf[index + 0x0e]) << 8)
	      + (((unsigned long) PartitionSecBuf[index + 0x0f]) << 0))
				* 1024 / Media.lbsize + p->Start - 1;
      p->Size = ((p->End) - (p->Start) + 1) / 1024 * Media.lbsize / 1024;

      p->OK = !0;
      if (memcmp ("Human68k", &PartitionSecBuf[index], 8))
	continue;

      if (Debug || Verbose)
	printf ("[Read PartitionHeadSec] ");
      ret = ReadSec (id, p->Start, 1, SecBuf);
      if (ret)
	Err_Read (id, ret);
      l = (((unsigned long) SecBuf[0x12]) << 8)
	+ (((unsigned long) SecBuf[0x13]) << 0);
      if (Buffers < l)
	Buffers = l;
      p->SecSize = l;
      l = (((unsigned long) SecBuf[0x14]) << 0);
      p->CluSec = l;
      l = (((unsigned long) SecBuf[0x16]) << 8)
	+ (((unsigned long) SecBuf[0x17]) << 0);
      p->FATSec = l;
      l = (((unsigned long) SecBuf[0x15]) << 0);
      p->FATCnt = l;
      l = (((unsigned long) SecBuf[0x18]) << 8)
	+ (((unsigned long) SecBuf[0x19]) << 0);
      p->RootDir = l;
      l = (((unsigned long) SecBuf[0x1a]) << 8)
	+ (((unsigned long) SecBuf[0x1b]) << 0);
      if (l == 0)
	{
	  l = (((unsigned long) SecBuf[0x1e]) << 24)
	    + (((unsigned long) SecBuf[0x1f]) << 16)
	    + (((unsigned long) SecBuf[0x20]) << 8)
	    + (((unsigned long) SecBuf[0x21]) << 0);
	}
      p->MaxSec = l;
      l = (((unsigned long) SecBuf[0x1c]) << 0);
      p->MediaID = l;
      l = (((unsigned long) SecBuf[0x1d]) << 0);
      p->FATSiz = l;
      p->HumanOK = !0;
    }

Quit:;
}


/***************************************************************/


void
GetPartitionSec_TOWNS (int id)
{
  int ret;

  PartitionSecBuf = malloc (Media.lbsize);
  if (NULL == PartitionSecBuf)
    Err_Memory ();

  if (Debug || Verbose)
    printf ("[Read PartitionSec] ");
  ret = ReadSec (id, 1, 1, PartitionSecBuf);
  if (ret)
    Err_Read (id, ret);
}


void
GetDiskStatus_TOWNS (int id)
{
  int index;
  int i;
  int ret;
  unsigned long l;

  SecBuf = malloc (Media.lbsize);
  if (NULL == SecBuf)
    Err_Memory ();

  GetPartitionSec_TOWNS (id);

  if (memcmp ("xm", &PartitionSecBuf[0], 6))
    goto Quit;

  printf ("Format		: FM-TOWNS \n");

  for (index = 0x20; index < Media.lbsize; index += 0x30)
    {
      struct PARTITION* p = &Partition[PartitionCnt++];

      if (0x20 == PartitionSecBuf[index + 0x10])
	continue;
      if (memcmp ("MS-DOS", &PartitionSecBuf[index + 0x10], 6))
	continue;

      for (i = 0; i < 16; i++)
	p->Name[i] = PartitionSecBuf[index + i + 0x20];
      p->Name[i] = '\0';

      p->BootFlag = ((!PartitionSecBuf[index + 0x00]) ? 0 : !0);
      p->ActiveFlag = ((!PartitionSecBuf[index + 0x01]) ? 0 : !0);
      p->Start = ((((unsigned long) PartitionSecBuf[index + 0x05]) << 24)
		+ (((unsigned long) PartitionSecBuf[index + 0x04]) << 16)
		+ (((unsigned long) PartitionSecBuf[index + 0x03]) << 8)
		+ (((unsigned long) PartitionSecBuf[index + 0x02]) << 0));
      p->End = ((((unsigned long) PartitionSecBuf[index + 0x09]) << 24)
	      + (((unsigned long) PartitionSecBuf[index + 0x08]) << 16)
	      + (((unsigned long) PartitionSecBuf[index + 0x07]) << 8)
	      + (((unsigned long) PartitionSecBuf[index + 0x06]) << 0))
	      + p->Start - 1;
      p->Size = ((p->End) - (p->Start) + 1) / 1024 * Media.lbsize / 1024;

      if (Debug || Verbose)
	printf ("[Read PartitionHeadSec] ");
      ret = ReadSec (id, p->Start, 1, SecBuf);
      if (ret)
	Err_Read (id, ret);
      l = (((unsigned long) SecBuf[12]) << 8)
	+ (((unsigned long) SecBuf[11]) << 0);
      if (Buffers < l)
	Buffers = l;
      p->SecSize = l;
      l = (((unsigned long) SecBuf[13]) << 0);
      p->CluSec = l;
      l = (((unsigned long) SecBuf[15]) << 8)
	+ (((unsigned long) SecBuf[14]) << 0);
      p->FATSec = l;
      l = (((unsigned long) SecBuf[16]) << 0);
      p->FATCnt = l;
      l = (((unsigned long) SecBuf[18]) << 8)
	+ (((unsigned long) SecBuf[17]) << 0);
      p->RootDir = l;
      l = (((unsigned long) SecBuf[20]) << 8)
	+ (((unsigned long) SecBuf[19]) << 0);
      if (l == 0)
	{
	  l = (((unsigned long) SecBuf[35]) << 24)
	    + (((unsigned long) SecBuf[34]) << 16)
	    + (((unsigned long) SecBuf[33]) << 8)
	    + (((unsigned long) SecBuf[32]) << 0);
	}
      p->MaxSec = l;
      l = (((unsigned long) SecBuf[21]) << 0);
      p->MediaID = l;
      l = (((unsigned long) SecBuf[23]) << 8)
	+ (((unsigned long) SecBuf[22]) << 0);
      p->FATSiz = l;
      p->OK = !0;
      p->HumanOK = !0;
    }
Quit:;
}


/***************************************************************/


void
GetDiskStatus_IBMMO (int id)
{
  int ret;
  unsigned long l;
  unsigned long secsize;
  unsigned long secnum;
  unsigned long clusec;

  SecBuf = malloc (Media.lbsize);
  if (NULL == SecBuf)
    Err_Memory ();

  l = (((unsigned long) BootSecBuf[15]) << 8)
    + (((unsigned long) BootSecBuf[14]) << 0);
  if (Debug || Verbose)
    printf ("[Read PartitionSec] ");
  ret = ReadSec (id, l, 1, SecBuf);
  if (ret)
    Err_Read (id, ret);
  if (memcmp ("\xff\xff\xff", &SecBuf[1], 3))
    goto Quit;

  printf ("Format		: IBM STANDARD-MO \n");

  {
    struct PARTITION* p = &Partition[PartitionCnt++];

    l = (((unsigned long) BootSecBuf[12]) << 8)
      + (((unsigned long) BootSecBuf[11]) << 0);
    secsize = l;
    l = (((unsigned long) BootSecBuf[20]) << 8)
      + (((unsigned long) BootSecBuf[19]) << 0);
    if (l == 0)
      {
	l = (((unsigned long) BootSecBuf[35]) << 24)
	  + (((unsigned long) BootSecBuf[34]) << 16)
	  + (((unsigned long) BootSecBuf[33]) << 8)
	  + (((unsigned long) BootSecBuf[32]) << 0);
      }
    secnum = l;
    l = (((unsigned long) BootSecBuf[13]) << 0);
    clusec = l;
    p->BootFlag = 0;
    p->ActiveFlag = !0;
    p->Start = 0;
    p->End = secnum;
    p->Size = secsize * secnum / 1024 / 1024;
    l = secsize;
    Buffers = l;
    if (Buffers < l)
      Buffers = l;
    p->SecSize = l;
    l = clusec;
    p->CluSec = l;
    l = (((unsigned long) BootSecBuf[15]) << 8)
      + (((unsigned long) BootSecBuf[14]) << 0);
    p->FATSec = l;
    l = (((unsigned long) BootSecBuf[16]) << 0);
    p->FATCnt = l;
    l = (((unsigned long) BootSecBuf[18]) << 8)
      + (((unsigned long) BootSecBuf[17]) << 0);
    p->RootDir = l;
    l = secnum;
    p->MaxSec = l;
    l = (((unsigned long) BootSecBuf[21]) << 0);
    p->MediaID = l;
    l = (((unsigned long) BootSecBuf[23]) << 8)
      + (((unsigned long) BootSecBuf[22]) << 0);
    p->FATSiz = l;
    p->OK = !0;
    p->HumanOK = !0;
  }

Quit:;
}


/***************************************************************/


void
GetPartitionSec_ZIP (int id)
{
  int ret;

  PartitionSecBuf = malloc (Media.lbsize);
  if (NULL == PartitionSecBuf)
    Err_Memory ();

  if (Debug || Verbose)
    printf ("[Read PartitionSec] ");
  ret = ReadSec (id, 0, 1, PartitionSecBuf);
  if (ret)
    Err_Read (id, ret);
}


void
GetDiskStatus_ZIP (int id)
{
  int index;
  int ret;
  unsigned long l;

  SecBuf = malloc (Media.lbsize);
  if (NULL == SecBuf)
    Err_Memory ();

  GetPartitionSec_ZIP (id);

  if (memcmp ("IPAR", &PartitionSecBuf[2], 4))
    goto Quit;

  printf ("Format		: ZIP Iomega \n");

  for (index = 0; index < 1; index++)
    {
      struct PARTITION* p = &Partition[PartitionCnt++];

      p->BootFlag = 0;
      p->ActiveFlag = !0;
      p->Start = ((((unsigned long) PartitionSecBuf[index + 0x1f9]) << 24)
		+ (((unsigned long) PartitionSecBuf[index + 0x1f8]) << 16)
		+ (((unsigned long) PartitionSecBuf[index + 0x1f7]) << 8)
		+ (((unsigned long) PartitionSecBuf[index + 0x1f6]) << 0));
      l = ((((unsigned long) PartitionSecBuf[index + 0x1fd]) << 24)
	 + (((unsigned long) PartitionSecBuf[index + 0x1fc]) << 16)
	 + (((unsigned long) PartitionSecBuf[index + 0x1fb]) << 8)
	 + (((unsigned long) PartitionSecBuf[index + 0x1fa]) << 0));
      p->Size = l * 512 / 1024 / 1024;
      p->End = p->Start + l;
      p->Name[0] = '\0';

      if (Debug || Verbose)
	printf ("[Read PartitionHeadSec] ");
      ret = ReadSec (id, p->Start, 1, SecBuf);
      if (ret)
	Err_Read (id, ret);
      l = (((unsigned long) SecBuf[12]) << 8)
	+ (((unsigned long) SecBuf[11]) << 0);
      if (Buffers < l)
	Buffers = l;
      p->SecSize = l;
      l = (((unsigned long) SecBuf[13]) << 0);
      p->CluSec = l;
      l = (((unsigned long) SecBuf[15]) << 8)
	+ (((unsigned long) SecBuf[14]) << 0);
      p->FATSec = l;
      l = (((unsigned long) SecBuf[16]) << 0);
      p->FATCnt = l;
      l = (((unsigned long) SecBuf[18]) << 8)
	+ (((unsigned long) SecBuf[17]) << 0);
      p->RootDir = l;
      l = (((unsigned long) SecBuf[20]) << 8)
	+ (((unsigned long) SecBuf[19]) << 0);
      if (l == 0)
	{
	  l = (((unsigned long) SecBuf[35]) << 24)
	    + (((unsigned long) SecBuf[34]) << 16)
	    + (((unsigned long) SecBuf[33]) << 8)
	    + (((unsigned long) SecBuf[32]) << 0);
	}
      p->MaxSec = l;
      l = (((unsigned long) SecBuf[21]) << 0);
      p->MediaID = l;
      l = (((unsigned long) SecBuf[23]) << 8)
	+ (((unsigned long) SecBuf[22]) << 0);
      p->FATSiz = l;
      p->OK = !0;
      p->HumanOK = !0;
    }
Quit:;
}


/***************************************************************/


void
GetPartitionSec_IBMFDISK (int id)
{
  int ret;

  IBMFDISKPartSec = 0;
  PartitionSecBuf = malloc (Media.lbsize);
  if (NULL == PartitionSecBuf)
    Err_Memory ();

  if (Debug || Verbose)
    printf ("[Read PartitionSec] ");
  ret = ReadSec (id, IBMFDISKPartSec, (SCSI_BLOCKSIZE ? 1 : 2), PartitionSecBuf);
  if (ret)
    Err_Read (id, ret);
}


void
GetDiskStatus_IBMFDISK (int id)
{
  int index;
  int i;
  int ret;
  unsigned long l;
  unsigned long ll;
  char ExPartition;

  ExPartition = 0;
  SecBuf = malloc (Media.lbsize);
  if (NULL == SecBuf)
    Err_Memory ();

  GetPartitionSec_IBMFDISK (id);

  index = 0;
  printf ("Format		: IBM-FDISK\n");
  while (IBMFDISKPartSec != 0xffffffff)
    {
      struct PARTITION* p = &Partition[PartitionCnt++];
      int NoDOS = 0;

      index++;
      for (i = 0; i < 16; i++)
	p->Name[i] = '\0';
      p->Name[i] = '\0';

      sprintf (p->Name, "(Not MS-DOS %d)", index);
      p->BootFlag = !0;
      p->ActiveFlag = !0;
    Retry_cylsize:;
      if (ExPartition)
	{
	  p->Start = ((((unsigned long) PartitionSecBuf[0x1c9]) << 24)
		    + (((unsigned long) PartitionSecBuf[0x1c8]) << 16)
		    + (((unsigned long) PartitionSecBuf[0x1c7]) << 8)
		    + (((unsigned long) PartitionSecBuf[0x1c6]) << 0))
		    + IBMFDISKPartSec;
	  p->End = ((((unsigned long) PartitionSecBuf[0x1cd]) << 24)
		  + (((unsigned long) PartitionSecBuf[0x1cc]) << 16)
		  + (((unsigned long) PartitionSecBuf[0x1cb]) << 8)
		  + (((unsigned long) PartitionSecBuf[0x1ca]) << 0))
		  + p->Start - 1;
	}
      else
	{
	  p->Start =
	      ((((unsigned long) PartitionSecBuf[(index - 1) * 0x10 + 0x1c9]) << 24)
	     + (((unsigned long) PartitionSecBuf[(index - 1) * 0x10 + 0x1c8]) << 16)
	     + (((unsigned long) PartitionSecBuf[(index - 1) * 0x10 + 0x1c7]) << 8)
	     + (((unsigned long) PartitionSecBuf[(index - 1) * 0x10 + 0x1c6]) << 0))
	     + IBMFDISKPartSec;
	  p->End = ((((unsigned long) PartitionSecBuf[(index - 1) * 0x10 + 0x1cd]) << 24)
		  + (((unsigned long) PartitionSecBuf[(index - 1) * 0x10 + 0x1cc]) << 16)
		  + (((unsigned long) PartitionSecBuf[(index - 1) * 0x10 + 0x1cb]) << 8)
		  + (((unsigned long) PartitionSecBuf[(index - 1) * 0x10 + 0x1ca]) << 0))
		  + p->Start - 1;
	  if (p->Start == 0)
	    {
	      IBMFDISKPartSec = 0xffffffff;
	      continue;
	    }
	}
      p->Size = ((p->End) - (p->Start) + 1) / 1024 * Media.lbsize / 1024;
      p->OK = !0;

      if (Debug || Verbose)
	printf ("[Read PartitionHeadSec %d] ", index);
      ret = ReadSec (id, p->Start, 1, SecBuf);
      if (ret)
	Err_Read (id, ret);
      if (SecBuf[0] != 0xeb)
	{
	  if ((Media.cylsize) == (Media.track * Media.head))
	    {
	      NoDOS = !0;
	      goto NoRetry;
	    }
	  if (Media.track * Media.head == 0)
	    {
	      NoDOS = !0;
	      goto NoRetry;
	    }
	  if (Media.cylsize != Media.track * Media.head)
	    {
	      Media.cylsize = Media.track * Media.head;
	      goto Retry_cylsize;
	    }
	  NoDOS = !0;
	}
    NoRetry:;
      if (ExPartition)
	{
	  if (PartitionSecBuf[0x1c2] & 0x80)
	    NoDOS = !0;
	}
      else
	{
	  if (PartitionSecBuf[(index - 1) * 0x10 + 0x1c2] & 0x80)
	    NoDOS = !0;
	}
      if ((index == 1) && (PartitionSecBuf[0x1d2] == 0x05))
	ExPartition = !0;

      if (ExPartition)
	{
	  if (Debug || Verbose)
	    printf ("[Read PartitionSec %d] ", index + 1);
	  ll = ((((unsigned long) PartitionSecBuf[0x1d9]) << 24)
	      + (((unsigned long) PartitionSecBuf[0x1d8]) << 16)
	      + (((unsigned long) PartitionSecBuf[0x1d7]) << 8)
	      + (((unsigned long) PartitionSecBuf[0x1d6]) << 0));
	  if (ll == 0)
	    IBMFDISKPartSec = 0xffffffff;
	  else
	    {
	      ll = ((((unsigned long) PartitionSecBuf[0x1c9]) << 24)
		  + (((unsigned long) PartitionSecBuf[0x1c8]) << 16)
		  + (((unsigned long) PartitionSecBuf[0x1c7]) << 8)
		  + (((unsigned long) PartitionSecBuf[0x1c6]) << 0))
		 + ((((unsigned long) PartitionSecBuf[0x1cd]) << 24)
		  + (((unsigned long) PartitionSecBuf[0x1cc]) << 16)
		  + (((unsigned long) PartitionSecBuf[0x1cb]) << 8)
		  + (((unsigned long) PartitionSecBuf[0x1ca]) << 0));
	      IBMFDISKPartSec += ll;
	      ret = ReadSec (id, IBMFDISKPartSec, (SCSI_BLOCKSIZE ? 1 : 2),
							PartitionSecBuf);
	      if (ret)
		Err_Read (id, ret);
	    }
	}
      if (NoDOS)
	continue;

      memcpy (p->OEMName, &SecBuf[3], 8);
      p->OEMName[8] = '\0';
      memcpy (p->FATName, &SecBuf[82], 8);
      p->FATName[8] = '\0';

      l = (((unsigned long) SecBuf[12]) << 8)
	+ (((unsigned long) SecBuf[11]) << 0);
      if (Buffers < l)
	Buffers = l;
      p->SecSize = l;
      l = (((unsigned long) SecBuf[13]) << 0);
      p->CluSec = l;
      l = (((unsigned long) SecBuf[15]) << 8)
	+ (((unsigned long) SecBuf[14]) << 0);
      p->FATSec = l;
      l = (((unsigned long) SecBuf[16]) << 0);
      p->FATCnt = l;
      l = (((unsigned long) SecBuf[18]) << 8)
	+ (((unsigned long) SecBuf[17]) << 0);
      p->RootDir = l;
      l = (((unsigned long) SecBuf[20]) << 8)
	+ (((unsigned long) SecBuf[19]) << 0);
      if (l == 0)
	{
	  l = (((unsigned long) SecBuf[35]) << 24)
	    + (((unsigned long) SecBuf[34]) << 16)
	    + (((unsigned long) SecBuf[33]) << 8)
	    + (((unsigned long) SecBuf[32]) << 0);
	}
      p->MaxSec = l;
      l = (((unsigned long) SecBuf[21]) << 0);
      p->MediaID = l;
      l = (((unsigned long) SecBuf[23]) << 8)
	+ (((unsigned long) SecBuf[22]) << 0);
      p->FATSiz = l;

      /* FAT32 */
      if (l == 0)
	{
	  l = (((unsigned long) SecBuf[39]) << 24)
	    + (((unsigned long) SecBuf[38]) << 16)
	    + (((unsigned long) SecBuf[37]) << 8)
	    + (((unsigned long) SecBuf[36]) << 0);
	}
      p->FATSiz = l;

      p->HumanOK = !0;
      sprintf (p->Name, "Partition %d", index);
    }
}


/***************************************************************/


void
GetDiskStatusCD_ISO9660 (int id)
{

  SecBuf = malloc (Media.lbsize);
  if (NULL == SecBuf)
    Err_Memory ();

  printf ("Format		: ISO-9660 CD-ROM\n");
}


/***************************************************************/


void
GetDiskStatusDVD_UDF (int id)
{

  SecBuf = malloc (Media.lbsize);
  if (NULL == SecBuf)
    Err_Memory ();

  printf ("Format		: Universal Disk Format\n");
}


/***************************************************************/


void
GetInitiatorInfo (int id)
{
  void* port;
  unsigned long portdata;
  int i;

  CPUNo = B_BPEEK ((unsigned char*) 0xcbc);
  VERNo = ROMVER ();

  SCSIPort = 0;
  SCSIRomLevel = TW_CALL (0x00a, id) & 0xffff;

  if (ReadyTWOSCSI)
    {
      i = TW_CALL (0x20a, id);
      SCSIPort = ((i & 0x10000) ? 1 : 2);
      SCSIRomLevel = i & 0xffff;
      goto OKCheckPort;
    }

  port = (unsigned char*) 0xea0020;
  if (BUS_ERR (port, port, 4) < 2)
    {
      portdata = B_LPEEK ((unsigned long*) (port + 0x26));
      if (portdata == 0x53494558)	/* 'SIEX' */
	{
	  SCSIPort = 1;
	  goto OKCheckPort;
	}
    }
  port = (unsigned char*) 0xfc0000;
  if (BUS_ERR (port, port, 4) < 2)
    {
      portdata = B_LPEEK ((unsigned long*) (port + 0x26));
      if (portdata == 0x5349494e)	/* 'SIIN' */
	{
	  SCSIPort = 2;
	  goto OKCheckPort;
	}
    }
OKCheckPort:;

}

void
PutInitiator (int id)
{
  GetInitiatorInfo (id & 7);

  printf ("Machine CPU	: 680%d0\n", CPUNo);
  printf ("Machine ROM	: ");
  printf ("%1x.%1x ", ((VERNo >> 24) & 0xff) / 0x10,
	  ((VERNo >> 24) & 0xff) % 0x10);
  printf ("%d%02x/%02x/%02x", 19 + (((VERNo >> 16) & 0xff) < 0x80),
		((VERNo >> 16) & 0xff), ((VERNo >> 8) & 0xff), ((VERNo >> 0) & 0xff));
  printf ("\n");
  printf ("Interface	: ");
  switch (SCSIPort)
    {
    case 1:
      if ((0x6411 <= SCSIRomLevel) && (SCSIRomLevel <= 0x641f))
	printf ("%s", "External SCSI Interface Board (Mach-2)");
      else
	printf ("%s", "External SCSI Interface Board");
      break;
    case 2:
      printf ("%s", "Internal SCSI Interface");
      break;
    default:
      printf ("%s", "SxSI on SASI Interface");
      break;
    }
  printf (" - ROM Level=%04X\n", SCSIRomLevel);

  if (ReadyTWOSCSI)
    {
      GetInitiatorInfo ((id & 7) + 8);
      printf ("Interface 2nd	: ");
      switch (SCSIPort)
	{
	case 1:
	  if ((0x6411 <= SCSIRomLevel) && (SCSIRomLevel <= 0x641f))
	    printf ("%s", "External SCSI Interface Board (Mach-2)");
	  else
	    printf ("%s", "External SCSI Interface Board");
	  break;
	case 2:
	  printf ("%s", "Internal SCSI Interface");
	  break;
	default:
	  printf ("%s", "SxSI on SASI Interface");
	  break;
	}
      printf (" - ROM Level=%04X\n", SCSIRomLevel);
    }
}


void
PutInitiator1 (int id)
{
  GetInitiatorInfo (id);

/*	     0123456789012345678901234567890123456789012345678901234567980123456789012345678901234567 */
/*  printf ("ID: Device                  RMB: Ver.   Capacity Sec: bender  product         revision\n"); */
/*  printf ("-----------------------------------------------------------------------------------------------\n"); */
  printf ("%2d: ", id & 0xffff);
  printf ("%-25s ", "Initiator");
  printf (" : ");

  switch (SCSIPort)
    {
    case 1:
      if ((0x6411 <= SCSIRomLevel) && (SCSIRomLevel <= 0x641f))
	printf ("%-13s(%04X): ", "SCSI Mach-2", SCSIRomLevel);
      else
	printf ("%-13s(%04X): ", "SCSI External", SCSIRomLevel);
      break;
    case 2:
      printf ("%-13s(%04X): ", "SCSI Internal", SCSIRomLevel);
      break;
    default:
      printf ("%-13s(%04X): ", "SxSI on SASI", SCSIRomLevel);
      break;
    }
  printf ("%s", "SHARP   ");
  printf ("X680%d0          ", CPUNo);
  printf ("%1x.%1x ", ((VERNo >> 24) & 0xff) / 0x10,
	  ((VERNo >> 24) & 0xff) % 0x10);
  printf ("%d%02x/%02x/%02x", 19 + (((VERNo >> 16) & 0xff) < 0x80),
		((VERNo >> 16) & 0xff), ((VERNo >> 8) & 0xff), ((VERNo >> 0) & 0xff));
  printf ("\n");
}


void
PutMediaStatus (int id)
{
  int lun = id >> 16;

  GetMediaStatus (id);

  if (lun)
    printf ("ID / LUN	: %d / %d\n", id & 0xffff, lun);
  else
    printf ("ID		: %d\n", id);
  printf ("Inquiry-Unique	: \"%s\"\n", &InquiryBuf[8]);
  printf ("		: (Lbender Lproduct        Lrevision)\n");
  printf ("Inquiry-Device	: %s\n", DeviceName[InquiryBuf[0] & 0x1f]);
  printf ("Inquiry-RMB	: %s\n",
	  ((InquiryBuf[1] & 0x80) ? "Removable" : "Unremovable"));
  printf ("Inquiry-Version	: ISO=%d ECMA=%d ANSI=%d (%s)\n",
	  (InquiryBuf[2] >> 6) & 0x03, (InquiryBuf[2] >> 3) & 0x07,
	  (InquiryBuf[2] >> 0) & 0x07, ANSIVersion[InquiryBuf[2] & 0x07]);
  printf ("Inquiry-Restype	: %s\n", ResType[InquiryBuf[3] & 0x0f]);
  if (ReadCap.size)
    {
      printf ("Block capacity	: $%lX (%luMBytes)\n", Media.lblock,
	      (Media.lblock / 1024) * Media.lbsize / 1024);
      printf ("Block size	: %lu\n", Media.lbsize);
    }
  else
    {
      printf ("Block capacity	: \n");
      printf ("Block size	: \n");
    }
  printf ("Cylinder size	: $%03lX (%s)\n", Media.cylsize,
	  (Media.okmodesense ? "by MODE SENSE" : "no MODE SENSE"));
  printf ("Sec/Track size	: $%03lX (%s)\n", Media.track,
	  (Media.okmodesense ? "by MODE SENSE" : "no MODE SENSE"));
  printf ("Number of Head	: %4ld (%s)\n", Media.head,
	  (Media.okmodesense ? "by MODE SENSE" : "no MODE SENSE"));
}


void
PutDiskStatus (int id)
{
  int i;
  int ret;
  int ibm = 0;
  char s[256];

  GetBootSec (id);

  if (!memcmp ("IPL1", &BootSecBuf[4], 4))
    {
      GetDiskStatus_PC98 (id);
      goto Print;
    }
  if (!memcmp ("X68SCSI1", &BootSecBuf[0], 8))
    {
      GetDiskStatus_X68K (id);
      goto Print;
    }
  if (!memcmp ("IPL4", &BootSecBuf[0], 4))
    {
      GetDiskStatus_TOWNS (id);
      goto Print;
    }
  if (!memcmp ("IPAR", &BootSecBuf[2], 4))
    {
      GetDiskStatus_ZIP (id);
      goto Print;
    }
  if ((0x55 == BootSecBuf[0x1fe]) & (0xaa == BootSecBuf[0x1ff]))
    {
      switch (BootSecBuf[0])
	{
	case 0xb8:
	case 0xfa:
	case 0x00:
	case 0x0e:
	  ibm = 1;
	  GetDiskStatus_IBMFDISK (id);
	  goto Print;
	case 0xeb:
	  ibm = 1;
	  GetDiskStatus_IBMMO (id);
	  goto Print;
	}
    }

  ret = ReadSec (id, 0x10, 1, BootSecBuf);
  if (ret)
    Err_Read (id, ret);
  if (memcmp ("\x00\x42\x45\x41\x30\x31\x01", &BootSecBuf[0], 7) == 0)
    {
      GetDiskStatusDVD_UDF (id);
      goto Quit;
    }
  if (memcmp ("\x01" "CD001", &BootSecBuf[0], 6) == 0)
    {
      ret = ReadSec (id, 0x12, 1, BootSecBuf);
      if (ret)
	Err_Read (id, ret);
      if (memcmp ("\x00\x42\x45\x41\x30\x31\x01", &BootSecBuf[0], 7) == 0)
	{
	  GetDiskStatusDVD_UDF (id);
	  goto Quit;
	}
    }

  printf ("Format		: Unrecognized\n");
  goto Quit;

Print:;
  for (i = 0; i < PartitionCnt; i++)
    {
      struct PARTITION* p = &Partition[i];

      if (!p->OK)
	continue;
      strcpy (s, "\"");
      strcat (s, p->Name);
      strcat (s, "\"");
      printf ("		  $%02X	", i);
      printf ("%-18s ", s);
      printf ("[%s ", ((p->BootFlag) ? "Boot  " : "Unboot"));
      printf ("%s] ", ((p->ActiveFlag) ? "Active" : "Sleep "));
      printf ("Start=$%08lX Size=%luMByte", p->Start, p->Size);
      printf ("\n");

      if (!p->HumanOK)
	continue;

      printf ("			 ");
      printf ("SecSize=$%04lX ", p->SecSize);
      printf ("CluSec=$%04lX ", p->CluSec);
      printf ("FATSec=$%04lX ", p->FATSec);
      printf ("FATCnt=$%04lX", p->FATCnt);
      printf ("\n			 ");
      printf ("RootDir=%5ld ", p->RootDir);
      printf ("MaxSec=$%08lX ", p->MaxSec);
      printf ("MediaID=$%02lX ", p->MediaID);
      printf ("FATSiz=$%04lX", p->FATSiz);
      if (ibm)
	{
	  printf ("\n			 ");
	  printf ("OEM=\"%8s\"", p->OEMName);
	  printf ("FAT=\"%8s\"", p->FATName);
	}
      printf ("\n");
    }

Quit:;
  free (BootSecBuf);
  free (PartitionSecBuf);
  free (SecBuf);
}


void
PutDiskStatusCD (int id)
{
  int ret;

  GetBootSecCD (id);

  if (!memcmp ("\x01" "CDROM", &BootSecBuf[0], 6))
    {
      GetDiskStatusCD_ISO9660 (id);
      goto Print;
    }
  if (memcmp ("\x01" "CD001", &BootSecBuf[0], 6) == 0)
    {
      ret = ReadSec (id, 0x12, 1, BootSecBuf);
      if (ret)
	Err_Read (id, ret);
      if (memcmp ("\x00\x42\x45\x41\x30\x31\x01", &BootSecBuf[0], 7) == 0)
	{
	  GetDiskStatusDVD_UDF (id);
	  goto Print;
	}
      GetDiskStatusCD_ISO9660 (id);
      goto Print;
    }
  ret = ReadSec (id, 0, 1, BootSecBuf);
  if (ret)
    Err_Read (id, ret);
  if ((0x55 == BootSecBuf[0x1fe]) && (0xaa == BootSecBuf[0x1ff]))
    {
      PutDiskStatus (id);
      goto Print;
    }
  printf ("Format		: Unrecognized\n");
  goto Quit;

Print:;

Quit:;
  free (BootSecBuf);
  free (PartitionSecBuf);
  free (SecBuf);
}


void
PutMediaStatus1 (int id)
{
  GetMediaStatus (id);
/*	     01234567890123456789012345678901234567890123456789012345679801234567890123456789012345678901234 */
/*  printf ("ID: Device                  RMB  Version Capacity Sect  bender  product         revision\n"); */
/*  printf ("-----------------------------------------------------------------------------------------------\n"); */
  printf ("%2d: ", id & 0xffff);
  printf ("%-25s ", DeviceName[InquiryBuf[0] & 0x1f]);
  printf ("%1d: ", ((InquiryBuf[1] & 0x80) ? 1 : 0));
  printf ("%-6s ", ANSIVersion[InquiryBuf[2] & 0x07]);
  if (ReadCap.size)
    {
      printf ("%5luMB ", (Media.lblock / 1024) * Media.lbsize / 1024);
      printf ("%4lu: ", Media.lbsize);
    }
  else
    printf ("%s", "            : ");

  if (Debug || Verbose)
    printf ("%s", &InquiryBuf[8]);
  else
    {
      InquiryBuf[8 + 41] = '\0';
      InquiryBuf[8 + 42] = '\0';
      printf ("%-41s", &InquiryBuf[8]);
    }
  printf ("\n");
}


void
PutUnreadyStatus1 (int id)
{
/*	     0123456789012345678901234567890123456789012345678901234567980123456789012345678901234567 */
/*  printf ("ID: Device                  RMB: Ver.   Capacity Sec: bender  product         revision\n"); */
/*  printf ("-----------------------------------------------------------------------------------------------\n"); */
  printf ("%2d: ", id & 0xffff);
  printf ("%s ", "-- Not Ready --");
  printf ("\n");
}


/***************************************************************/


void
PutDiskDump (int id, unsigned long start, unsigned long count)
{
  unsigned long l;
  unsigned char* SecBuf;

  SecBuf = malloc (Media.lbsize);
  if (NULL == SecBuf)
    Err_Memory ();

  Verbose = !0;
  Debug = !0;

  for (l = 0; l < count; l++)
    ReadSec (id, start + l, 1, SecBuf);
}


/***************************************************************/


int
PrintTarget (int id)
{
  int ret;
  int retTestUnit;

  PutInitiator (id);

  if (ReqResetBus)
    SetResetBus ();

#if 0
  ret = SetStartStop (id);
#endif
  ret = SetRezeroUnit (id);
  if (ret < 0)
    goto NGRezeroUnit;
  retTestUnit = GetTestUnit (id);
  ret = GetInquiry (id);
  if (ret)
    goto NGInquiry;

  ReadCap.size = 0;
  ReadCap.block = 0;
  switch (InquiryBuf[0] & 0x1f)
    {
    case 0x02:			/* Printer Device */
    case 0x03:			/* Processor Device */
    case 0x06:			/* Scanner Device */
    case 0x08:			/* Disk Changer Device */
    case 0x09:			/* Communication Device */
      break;
    default:
      ret = GetReadCapacity (id);
      if ((!retTestUnit) && (ret))
	goto NGReadCapacity;
    }

  PutMediaStatus (id);

  if (!retTestUnit)
    {
      if (DumpData)
	PutDiskDump (id, DumpStart, DumpCount);
      else
	{
	  switch (InquiryBuf[0] & 0x1f)
	    {
	    case 0x05:		/* CD-ROM Device */
	      PutDiskStatusCD (id);
	      break;
	    default:
	      PutDiskStatus (id);
	      break;
	    }
	}
    }

/* OK:; */
  return 0;
NG:;
  return 1;

NGRezeroUnit:;
NGTestUnit:;
  if (InitiatorID == (id & 7))
    {
      printferr ("ID %d ͖{̃CjVG[^łB\n", id & 0xffff);
      goto NG;
    }
  printferr ("ID%d ͏łĂ܂(%X)B\n", id & 0xffff, ret);
  goto NG;

NGInquiry:;
  if (retTestUnit)
    goto NGTestUnit;
  printferr ("ID%d  INQUIRY 삵܂(%X)B\n", id & 0xffff,
	     ret);
  goto NG;

NGReadCapacity:;
  if (retTestUnit)
    goto NGTestUnit;
  printferr ("ID%d  READ CAPACITY 삵܂(%X)B\n",
	     id & 0xffff, ret);
  goto NG;

}


/***************************************************************/


int
PrintAllTarget (void)
{
  int i;
  int ret;
  int retTestUnit;
  int retReadCapacity;

/*         0123456789012345678901234567890123456789012345678901234567980123456789012345678901234567 */
  printf ("ID: Device                  RMB: Ver.   Capacity Sec: bender  product         revision\n");
  printf ("-----------------------------------------------------------------------------------------------\n");

  if (ReqResetBus)
    SetResetBus ();
  for (i = 0; i <= MaxID; i++)
    {
      int lun;
      for (lun = 0; lun <= MaxLUN; lun++)
	{
	  int id = i + (lun << 16);
#if 0
	  ret = SetStartStop (id);
#endif
	  ret = SetRezeroUnit (id);
	  if (ret < 0)
	    goto NGRezeroUnit;
	  retTestUnit = GetTestUnit (id);
	  ret = GetInquiry (id);
	  if (ret)
	    goto NG;

	  switch (InquiryBuf[0] & 0x1f)
	    {
	    case 0x02:		/* Printer Device */
	    case 0x03:		/* Processor Device */
	    case 0x06:		/* Scanner Device */
	    case 0x08:		/* Disk Changer Device */
	    case 0x09:		/* Communication Device */
	      break;
	    default:
	      ret = GetReadCapacity (id);
	      retReadCapacity = ret;
	    }

	  PutMediaStatus1 (id);

	  continue;
	NG:;
	  if (InitiatorID == (id & 7))
	    goto INITIATOR;
	  if (Verbose || Debug)
	    {
	      if (retTestUnit)
		{
		  printferr ("ID%d ͏łĂ܂(%X)B\n",
			     id & 0xffff, retTestUnit);
		}
	      if (ret)
		{
		  printferr ("ID%d  INQUIRY 삵܂(%X)B\n",
			     id & 0xffff, ret);
		}
	    }
	  PutUnreadyStatus1 (id);
	  continue;
	NGRezeroUnit:;
	  if (InitiatorID == (id & 7))
	    goto INITIATOR;
	  if (Verbose || Debug)
	    {
	      printferr ("ID%d ͏łĂ܂(%X)B\n", id & 0xffff,
			 ret);
	    }
	  PutUnreadyStatus1 (id);
	  continue;
	INITIATOR:;
	  PutInitiator1 (id);
	  continue;
	}
    }

/*	    0123456789012345678901234567890123456789012345678901234567980123456789012345678901234567 */
/*  printf("ID: Device                  RMB: Ver.   Capacity Sec: bender  product         revision\n"); */
  printf ("-----------------------------------------------------------------------------------------------\n");

  return 0;
}


/***************************************************************/


void
InitializeKeepSUSIE (void)
{
  int sp;
  unsigned char* devp;
  long l;

  sp = SUPER (0);

  devp = (unsigned char*) 0x6800;
  for (l = 0; l < (0x10000 - 0x6800) / 2; l++)
    {
      if (!memcmp (devp, "NUL ", 4))
	goto Found;
      devp += 2;
    }
  goto Quit;

Found:;
  devp -= 0x0e;

  do
    {
      if (!memcmp (devp + 0x0e, "\x01*SUSIE*", 8))
	{
	  register int _d0 asm ("d0") = 1;
	  register int _d1 asm ("d1") = 0;
	  register unsigned char* _a1 asm ("a1") = devp + 22;
	  __asm ("jsr	(a1)\n"
		: /* no output */
		: "d" (_d0), "d" (_d1), "a" (_a1)
		: "d0", "d1", "a1");
	}
      devp = *((unsigned char**) devp);
    }
  while (((long) devp) >= 0);

Quit:;
  if (sp >= 0)
    SUPER (sp);
}


/***************************************************************/


/************************************************************************
 *									*
 *	AnalyzeArg()							*
 *									*
 *	argv[]͂AIvVXCb`FďB	*
 *	IvVXCb`łȂ̂Argv[]Ɋi[B		*
 *	͂I_ŁAi[̌ArgcɊi[B	*
 *	Argv[0]̓vZXi[AArgv[1]i[B	*
 *	Argv̍ő̓}NMaxArgvɐݒ肳ĂB			*
 *									*
 *	IvVXCb`Ƃ"-"Ŏn܂SĂ̂̕ƂwB	*
 *	ϐSLASH"/"łȂ΁A"-""/"ł悢B		*
 *									*
 *	in	int	argc	R}hCp[^̐		*
 *				main()󂯎̂nB		*
 *		char	argv[]	R}hCp[^		*
 *				main()󂯎̂nB		*
 *									*
 *	out	Ȃ							*
 *									*
 ************************************************************************/

#define	SkipComma(p)	{ if (',' == *(p)) p++; }

static void
AnalyzeArg (int argc, char* argv[])
{
  int slash = !0;

  /* ϐ SLASH QƂ */
  {
    char* p;

    p = getenv ("SLASH");
    if (p != NULL)
      {
	if ('/' == *(p))
	  {
	    slash = 0;
	  }
      }
  }

  /* Argv[0]ɃvZXݒ肷 */
  {
    char* p1;
    char* p2 = NULL;

    p1 = strtok (argv[0], "\\/");
    while (NULL != p1)
      {
	p2 = p1;
	p1 = strtok (NULL, "\\/");
      }
    Argv[Argc++] = p2;
    ProcName = p2;
  }

  while ((--argc) > 0)
    {
      char* p;
      char c;
      char* p1;
      long l;

      p = *(++argv);
      p1 = p;
      c = *(p++);
      if (('-' == c) || ((slash && ('/' == c))))
	{
	  switch (*(p++))
	    {

	    case '?':
	      PrintHelp ();
	      break;

	    case 'd':
	    case 'D':
	      /* -D */
	      DumpData = !0;
	      DumpStart = 0;
	      DumpCount = 1;
	      if (*(p) != '\0')
		{
		  if (*(p) == '$')
		    {
		      p++;
		      l = strtol (p, &p, 16);
		    }
		  else
		    {
		      l = strtol (p, &p, 0);
		    }
		  DumpStart = (unsigned long) l;
		  if (*(p) == ',')
		    {
		      p++;
		      if (*(p) == '$')
			{
			  p++;
			  l = strtol (p, &p, 16);
			}
		      else
			{
			  l = strtol (p, &p, 0);
			}
		      DumpCount = (unsigned long) l;
		    }
		}
	      break;

	    case 'l':
	    case 'L':
	      /* -L */
	      MaxLUN = 7;
	      break;

	    case 'r':
	    case 'R':
	      /* -R */
	      switch (*p)
		{
		case 's':
		case 'S':
		  /* -RS */
		  p++;
		  ReqResetBus = !0;
		  break;
		}
	      break;

	    case 'v':
	    case 'V':
	      /* -V */
	      Verbose = !0;
	      break;

	    case '$':
	      /* -$ */
	      Debug = !0;
	      break;

	    default:
	      Err_Option (p1);

	    }
	  if ('\0' != *(p))
	    {
	      Err_Option (p1);
	    }
	}
      else
	{
	  if (Argc > MaxArgc)
	    Err_OverArg (p1);
	  Argv[Argc++] = p1;
	}
    }

}


/***************************************************************/


int
main (int argc, char* argv[])
{
  int id = 0;
  int ret;
  char* s;

  InitializeKeepSUSIE ();

  printf ("AskSCSI: Ask SCSI status : V1.14A+1 : Copyright 1994-96 GORRY.\n");

  if (TWOSCHK () == -2)
    {
      ReadyTWOSCSI = 1;
      MaxID = MAXIDCHK ();
      if (MaxID < 0)
	MaxID = 15;
    }

  InitiatorID = B_BPEEK ((unsigned char*) 0xed0070) & 0x07;

  AnalyzeArg (argc, argv);

  if (Argc > 1)
    {
      id = strtol (Argv[1], &s, 10);
      if (*s || (id < 0) || (id > MaxID))
	PrintHelp ();
      if (Argc > 2)
	{
	  int lun = strtol (Argv[2], &s, 10);
	  if (*s || (lun < 0) || (lun > 7))
	    PrintHelp ();
	  id += (lun << 16);
	}
      ret = PrintTarget (id);
    }
  else
    {
      if (DumpData)
	PrintHelp ();
      ret = PrintAllTarget ();
    }

  if (ret)
    return EXIT_FAILURE;

  if (!DumpData)
    {
      if (B_WPEEK ((unsigned short*) 0x1c70) < Buffers)
	{
	  printferr ("******************************************************************************\n"
		     "* CONFIG.SYS  \"BUFFERS=\"̑Qp[^ %lu ȏɂKv܂B *\n"
		     "******************************************************************************\x07\n\n",
			Buffers);
	}
      if (Verbose || Debug)
	{
	  if (Media.needSUSIE98COption)
	    {
	      printferr ("******************************************************************************\n"
			 "* SUSIE.XŔFȂꍇ \"-98:C%lu\" tĂ݂ĂB	     *\n"
			 "******************************************************************************\x07\n\n",
				Media.cylsize);
	    }
	}
    }

  return EXIT_SUCCESS;
}


/***************************************************************/
